# encoding: utf-8

require 'monitor'
require 'one_apm/transaction/event_analytic_sample'
require 'one_apm/transaction/event_analytic_data'

class OneApm::Collector::TransactionEventAggregator
  include OneApm::Coerce, MonitorMixin

  # The type field of the sample
  OA_SAMPLE_TYPE                    = 'Transaction'.freeze

  # Strings for static keys of the sample structure
  OA_TYPE_KEY                       = 'type'.freeze
  OA_TIMESTAMP_KEY                  = 'timestamp'.freeze
  OA_NAME_KEY                       = 'name'.freeze
  OA_DURATION_KEY                   = 'duration'.freeze
  OA_HTTP_RESPONSE_CODE_KEY         = 'httpResponseCode'.freeze
  OA_GUID_KEY                       = 'bw.guid'.freeze
  OA_REFERRING_TRANSACTION_GUID_KEY = 'bw.referringTransactionGuid'.freeze
  OA_CAT_TRIP_ID_KEY                = 'bw.tripId'.freeze
  OA_CAT_PATH_HASH_KEY              = 'bw.pathHash'.freeze
  OA_CAT_REFERRING_PATH_HASH_KEY    = 'bw.referringPathHash'.freeze
  OA_CAT_ALTERNATE_PATH_HASHES_KEY  = 'bw.alternatePathHashes'.freeze
  OA_APDEX_PERF_ZONE_KEY            = 'bw.apdexPerfZone'.freeze
  OA_SYNTHETICS_RESOURCE_ID_KEY     = "bw.syntheticsResourceId".freeze
  OA_SYNTHETICS_JOB_ID_KEY          = "bw.syntheticsJobId".freeze
  OA_SYNTHETICS_MONITOR_ID_KEY      = "bw.syntheticsMonitorId".freeze

  def initialize(event_listener)
    super()

    @enabled       = false
    @notified_full = false

    @samples            = ::OneApm::Agent::SampledBuffer.new(OneApm::Manager.config[:'analytics_events.max_samples_stored'])
    @synthetics_samples = ::OneApm::Agent::SyntheticsEventBuffer.new(OneApm::Manager.config[:'synthetics.events_limit'])

    event_listener.subscribe(:transaction_finished, &method(:on_transaction_finished))
    self.register_config_callbacks
  end

  # Fetch a copy of the sampler's gathered samples. (Synchronized)
  def samples
    self.synchronize { @samples.to_a.concat(@synthetics_samples.to_a) }
  end

  def reset!
    sample_count, request_count, synthetics_dropped = 0
    old_samples = nil

    self.synchronize do
      sample_count = @samples.size
      request_count = @samples.num_seen

      synthetics_dropped = @synthetics_samples.num_dropped

      old_samples = @samples.to_a + @synthetics_samples.to_a
      @samples.reset!
      @synthetics_samples.reset!

      @notified_full = false
    end

    [old_samples, sample_count, request_count, synthetics_dropped]
  end

  # Clear any existing samples, reset the last sample time, and return the
  # previous set of samples. (Synchronized)
  def harvest!
    old_samples, sample_count, request_count, synthetics_dropped = reset!
    record_sampling_rate(request_count, sample_count) if @enabled
    record_dropped_synthetics(synthetics_dropped)
    # return old_samples.map(&:to_collector_array) if old_samples.respond_to?(:to_collector_array)
    old_samples
  end

  # Merge samples back into the buffer, for example after a failed
  # transmission to the collector. (Synchronized)
  def merge!(old_samples)
    self.synchronize do
      old_samples.each do |s|
        if s.respond_to?(:spec_name)
          append_event(s)
        else
          @samples.append_event(s)
        end
      end
    end
  end

  def record_sampling_rate(request_count, sample_count)
    request_count_lifetime = @samples.seen_lifetime
    sample_count_lifetime = @samples.captured_lifetime
    OneApm::Manager.logger.debug("Sampled %d / %d (%.1f %%) requests this cycle, %d / %d (%.1f %%) since startup" % [
      sample_count,
      request_count,
      (sample_count.to_f / request_count * 100.0),
      sample_count_lifetime,
      request_count_lifetime,
      (sample_count_lifetime.to_f / request_count_lifetime * 100.0)
    ])

    engine = OneApm::Manager.agent.stats_engine
    engine.tl_record_supportability_metric_count("TransactionEventAggregator/requests", request_count)
    engine.tl_record_supportability_metric_count("TransactionEventAggregator/samples", sample_count)
  end

  def record_dropped_synthetics(synthetics_dropped)
    return unless synthetics_dropped > 0

    OneApm::Manager.logger.debug("Synthetics transaction event limit (#{@samples.capacity}) reached. Further synthetics events this harvest period dropped.")

    engine = OneApm::Manager.agent.stats_engine
    engine.tl_record_supportability_metric_count("TransactionEventAggregator/synthetics_events_dropped", synthetics_dropped)
  end

  def register_config_callbacks
    OneApm::Manager.config.register_callback(:'analytics_events.max_samples_stored') do |max_samples|
      OneApm::Manager.logger.debug "TransactionEventAggregator max_samples set to #{max_samples}"
      self.synchronize { @samples.capacity = max_samples }
    end

    OneApm::Manager.config.register_callback(:'synthetics.events_limit') do |max_samples|
      OneApm::Manager.logger.debug "TransactionEventAggregator limit for synthetics events set to #{max_samples}"
      self.synchronize { @synthetics_samples.capacity = max_samples }
    end

    OneApm::Manager.config.register_callback(:'analytics_events.enabled') do |enabled|
      @enabled = enabled
    end
  end

  def notify_full
    OneApm::Manager.logger.debug "Transaction event capacity of #{@samples.capacity} reached, beginning sampling"
    @notified_full = true
  end

  # Event handler for the :transaction_finished event.
  def on_transaction_finished(payload)
    return unless @enabled
    main_event = create_main_event(payload)
    custom_params = create_custom_parameters(payload)
    sub_events = create_sub_events(main_event, payload)
    self.synchronize do 
       [main_event].concat(sub_events).each do |event|
          append_event(event) 
        end
     end
    notify_full if !@notified_full && @samples.full?
  end


  # Event handler for the :transaction_finished event.
  def on_cross_app_transaction_finished(payload)
    return unless @enabled
    main_event = create_main_event(payload)
    self.synchronize do
      append_event(main_event)
    end
    notify_full if !@notified_full && @samples.full?
  end

  def append_event(event)
    same_spec_for_sample = @samples.select{|sample| sample.spec_name == event.spec_name}.first
    return @samples.append(event) if same_spec_for_sample.nil?
    same_spec_for_sample.event_analytic_data.concat event.event_analytic_data
  end

  # see http://wiki.oneapm.me/display/SAAS/Ai-+Entrance
  def create_main_event(payload)
    OneApm::EventAnalyticSample.new(payload)
  end

  def create_sub_events main_event, payload
     payload = payload.dup
     payload[:scope] = payload[:name]
     payload[:referring_transaction_guid] = main_event.guid
     payload[:request_url] = ''

     sub_event_samples = []
     return sub_event_samples unless payload[:metrics]
     payload[:metrics].each_scoped do |metric_name, status|
      next if metric_name =~ /^Nested|View|External/
      payload[:name] = metric_name
      payload[:guid] = OneApm::Helper.generate_guid
      payload[:call_count] = status.call_count
      payload[:duration] = status.total_call_time
      sample = OneApm::EventAnalyticSample.new(payload)
      sub_event_samples << sample
    end
    sub_event_samples
  end

  def error_times(txn_metrics)
    return txn_metrics[OA_ERROR_ALL_KEY].call_count if txn_metrics.has_key?(OA_ERROR_ALL_KEY) rescue 0
    return 0
  end

  def append_http_response_code(sample, payload)
    unless OneApm::Manager.config[:disable_rack_middleware]
      optionally_append(OA_HTTP_RESPONSE_CODE_KEY, :http_response_code, sample, payload)
    end
  end

  def append_cat_alternate_path_hashes(sample, payload)
    if payload.include?(:cat_alternate_path_hashes)
      sample[OA_CAT_ALTERNATE_PATH_HASHES_KEY] = payload[:cat_alternate_path_hashes].sort.join(',')
    end
  end

  def optionally_append(sample_key, payload_key, sample, payload)
    if payload.include?(payload_key)
      sample[sample_key] = string(payload[payload_key])
    end
  end

  def create_custom_parameters(payload)
    custom_params = {}
    if OneApm::Manager.config[:'analytics_events.capture_attributes']
      custom_params.merge!(event_params(payload[:custom_params] || {}))
    end
    custom_params
  end

end
